package mpi.eudico.client.annotator.commands;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import nl.mpi.lexan.annotyzers.helpers.Position;
import nl.mpi.lexan.annotyzers.helpers.Suggestion;
import nl.mpi.lexan.annotyzers.helpers.SuggestionSet;
import mpi.eudico.server.corpora.clomimpl.abstr.AbstractAnnotation;
import mpi.eudico.server.corpora.clomimpl.abstr.TierImpl;
import mpi.eudico.server.corpora.clomimpl.abstr.TranscriptionImpl;
import mpi.eudico.server.corpora.clomimpl.type.Constraint;
import mpi.eudico.server.corpora.clomimpl.type.LinguisticType;

/**
 * An undoable command that creates annotations from a set of suggestions.
 * If the annotations are created on a depending tier, first existing annotations are
 * deleted.
 * 
 * @author Han Sloetjes
 */
public class AnnotationsFromSuggestionSetCommand implements UndoableCommand {
	private String name;
	private TranscriptionImpl transcription;
	
	/**
	 * Constructor accepting the name.
	 * 
	 * @param name the name/id of the command
	 */
	public AnnotationsFromSuggestionSetCommand(String name) {
		super();
		this.name = name;
	}

	/**
	 * @param receiver the transcription object
	 * @param arguments arguments[0] = SuggestionSet
	 */
	@Override
	public void execute(Object receiver, Object[] arguments) {
		transcription = (TranscriptionImpl) receiver;
		
		SuggestionSet sugSet = (SuggestionSet) arguments[0];
		// get the "top level" suggestions
		List<Suggestion> topSugs = sugSet.suggestions;
		if (topSugs != null && topSugs.size() > 0) {
			Suggestion first = topSugs.get(0);
			Position pos = first.position;
			
			TierImpl topChTier = (TierImpl) transcription.getTierWithId(pos.tierId);
			if (topChTier != null) {
				TierImpl topParent = (TierImpl) topChTier.getParentTier();
				if (topParent != null) {
					// the suggestions are for depending tiers
					AbstractAnnotation aaPar = (AbstractAnnotation) topParent.getAnnotationAtTime(
							(pos.beginTime + pos.endTime) / 2);
					if (aaPar != null) {
						List children = aaPar.getChildrenOnTier(topChTier);
						if (children.size() > 0) {
							// store for undo and remove
							for (int i = 0; i < children.size(); i++) {
								AbstractAnnotation aa = (AbstractAnnotation) children.get(i);
								// TODO for undo
								topChTier.removeAnnotation(aa);
							}
						}
						// could create annotation data records and use AnnotationRecreator to 
						// create the annotations
						
					}
				} else {
					// the suggestions are for a toplevel tier
				}
				// create a single Suggestion from the set so that the same method can be applied
				// for creating annotations
				Suggestion topSug = new Suggestion();
				topSug.children = sugSet.suggestions;
				
				Map<Suggestion, AbstractAnnotation> sugToAnnMap = 
						new HashMap<Suggestion, AbstractAnnotation>();
				sugToAnnMap.put(topSug, null);
				createAnnotationsForSuggestions(sugToAnnMap);
				
			} else {
				// no show: nothing to delete, nothing to create
			}
		}
	}
	
	/**
	 * Create child annotations for the child-suggestions.
	 * 
	 * @param sugAnnMap a mapping of already processed suggestions and the corresponding annotations
	 */
	private void createAnnotationsForSuggestions(Map<Suggestion, AbstractAnnotation> sugAnnMap) {
		if (sugAnnMap.size() == 0) {
			return;
		}
		// hier... check
		Iterator<Suggestion> suggestIter = sugAnnMap.keySet().iterator();
		
		while (suggestIter.hasNext()) {
			Suggestion nextSuggestion = suggestIter.next();
			if (nextSuggestion.children == null || nextSuggestion.children.size() == 0) {
				continue;
			}
			 
			Map<String, List<Suggestion>> groupedChildren = groupPerTier(nextSuggestion.children);
			if (groupedChildren == null) {
				continue;
			}
			AbstractAnnotation parAnn = sugAnnMap.get(nextSuggestion);
			// iterate over the grouped children and then process the children per group
			Iterator<String> childIter = groupedChildren.keySet().iterator();
			List<Suggestion> sugsPerTier = null;
			
			while (childIter.hasNext()) {
				String tierName = childIter.next();
				sugsPerTier = groupedChildren.get(tierName);
			
				int numSugs = sugsPerTier.size();
				
				long bt = -1;
				long et = -1;
				if (parAnn != null) {
					bt = parAnn.getBeginTimeBoundary();
					et = parAnn.getEndTimeBoundary();
				} else {
					bt = sugsPerTier.get(0).position.beginTime;
					et = sugsPerTier.get(numSugs - 1).position.endTime;
				}
				
				long perAnn = (et - bt) / numSugs;
				
				AbstractAnnotation curAnn = null;
				AbstractAnnotation prevAnn = null;
				Suggestion curSug;
				Map<Suggestion, AbstractAnnotation> createdAnns = 
						new HashMap<Suggestion, AbstractAnnotation>();
				
				for (int i = 0; i < numSugs; i++) {
					curSug = sugsPerTier.get(i);
					curAnn = null;
					TierImpl t = (TierImpl) transcription.getTierWithId(tierName);
					if (t != null) {
						// check type of tier
						LinguisticType lt = t.getLinguisticType();
						Constraint con = lt.getConstraints();
						
						if (con == null || con.getStereoType() == Constraint.INCLUDED_IN) {
//							curAnn = (AbstractAnnotation) t.createAnnotation(curSug.position.beginTime + (i * perAnn), 
//								curSug.position.beginTime + ((i + 1) * perAnn));
							curAnn = (AbstractAnnotation) t.createAnnotation(bt + (i * perAnn), bt + ((i + 1) * perAnn));
						} else if (con.getStereoType() == Constraint.TIME_SUBDIVISION) {
							if (prevAnn != null) {
								curAnn = (AbstractAnnotation) t.createAnnotationAfter(prevAnn);
							} else {
//								curAnn = (AbstractAnnotation) t.createAnnotation(curSug.position.beginTime + (i * perAnn), 
//										curSug.position.beginTime + ((i + 1) * perAnn));
								curAnn = (AbstractAnnotation) t.createAnnotation(bt + (i * perAnn), 
										bt + ((i + 1) * perAnn));
							}
						} else if (con.getStereoType() == Constraint.SYMBOLIC_SUBDIVISION) {
							if (prevAnn != null) {
								curAnn = (AbstractAnnotation) t.createAnnotationAfter(prevAnn);
							} else {
								//long mid = (curSug.position.beginTime + curSug.position.endTime) / 2;
								long mid = (bt + et) / 2;
								curAnn = (AbstractAnnotation) t.createAnnotation(mid, mid);
							}
						} else if (con.getStereoType() == Constraint.SYMBOLIC_ASSOCIATION) {
							// should we assume multiple sym. associated annotations/suggestions at the same level 
							if (prevAnn == null) {
								long mid = (bt + et) / 2;
								curAnn = (AbstractAnnotation) t.createAnnotation(mid, mid);
							}
						}
						
						if (curAnn != null) {
							curAnn.setValue(curSug.content);
							if (curSug.children != null && curSug.children.size() > 0) {
								createdAnns.put(curSug, curAnn);
							}
						}
					}
					
					prevAnn = curAnn;
				}
				// create children
				createAnnotationsForSuggestions(createdAnns);
			}
		}
		
	}

	/**
	 * Groups the suggestions in the list per tier and returns a map of tier name
	 * to suggestions mappings.
	 * 
	 * @param suggestions a list of suggestions that might be for different tiers
	 * 
	 * @return a map of tier name to suggestions mappings
	 */
	private Map<String, List<Suggestion>> groupPerTier(List<Suggestion> suggestions) {
		if (suggestions == null || suggestions.size() == 0) {
			return null;
		}
		
		Map<String, List<Suggestion>> groupedSugs = new HashMap<String, List<Suggestion>>(6);
		
		for (Suggestion sug : suggestions) {
			if (groupedSugs.containsKey(sug.position.tierId)) {
				groupedSugs.get(sug.position.tierId).add(sug);
			} else {
				List<Suggestion> sugGroup = new ArrayList<Suggestion>(6);
				sugGroup.add(sug);
				groupedSugs.put(sug.position.tierId, sugGroup);
			}
		}
		
		return groupedSugs;
	}
	
	
	@Override
	public String getName() {
		return name;
	}

	@Override
	public void undo() {
		// TODO Auto-generated method stub

	}

	@Override
	public void redo() {
		// TODO Auto-generated method stub

	}

}
