package mpi.eudico.client.annotator.commands;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.nio.charset.UnsupportedCharsetException;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.SortedSet;
import java.util.TreeSet;

import mpi.eudico.client.annotator.interannotator.AnnotatorCompareUtil2;
import mpi.eudico.client.annotator.interannotator.CompareCombi;
import mpi.eudico.client.annotator.interannotator.CompareConstants;
import mpi.eudico.client.annotator.interannotator.CompareUnit;
import mpi.eudico.client.annotator.interannotator.modkappa.AgreementTable;
import mpi.eudico.client.annotator.interannotator.modkappa.CKappa;
import mpi.eudico.client.annotator.interannotator.modkappa.KappaCompareCombi;
import mpi.eudico.client.annotator.interannotator.modkappa.TwoSquareTable;
import mpi.eudico.server.corpora.clom.AnnotationCore;

/**
 * A command that compares the annotations of two tiers by applying the 
 * Holle/Rein modified kappa algorithm.  
 * 
 */
public class CompareAnnotationModKappaCommand extends AbstractCompareCommand {
	private final String UNMATCHED = "Unmatched";
	List<KappaCompareCombi> modCompareCombis;
	private KappaCompareCombi globalCombi;
	/**
	 * Constructor.
	 * 
	 * @param theName
	 */
	public CompareAnnotationModKappaCommand(String theName) {
		super(theName);
	}

	/**
	 * Override by returning a list of the extended KappaCompareCombi objects
	 */
	@Override
	public List<CompareCombi> getCompareSegments() {
		if (modCompareCombis != null) {
			return new ArrayList<CompareCombi>(modCompareCombis);
		}
		return super.getCompareSegments();
	}

	/**
	 * The actual calculation, applies the algorithm to the compare combinations and
	 * stores intermediate results per annotation (= coding) value.
	 */
	@Override
	protected void calculateAgreement() {
		// in the super class checks have been performed and combinations of tiers have been
		// created. Start the calculations right away.
		if (compareSegments.size() == 0) {
			logErrorAndInterrupt("There are no tier pairs, nothing to calculate.");
			return;
		}
		// should be 51% or more, default is 60%
		double minimalOverlapPercentage = 0.6d; 
		Object prefOPObject = compareProperties.get(CompareConstants.OVERLAP_PERCENTAGE);
		if (prefOPObject instanceof Double) {
			minimalOverlapPercentage = (Double) prefOPObject;
		}
		boolean perFileOutput = false;
		Object prefPerFileObject = compareProperties.get(CompareConstants.OUTPUT_PER_FILE);
		if (prefPerFileObject instanceof Boolean) {
			perFileOutput = (Boolean) prefPerFileObject;
		}
		
		int combiCount = 0;
		modCompareCombis = new ArrayList<KappaCompareCombi>();
		// starting at an arbitrary 30%
		float perCombi = 50f / compareSegments.size();
		AnnotatorCompareUtil2 compareUtil = new AnnotatorCompareUtil2();
		progressUpdate((int) curProgress, "Building agreement tables, per file...");
		
		Map<AnnotationCore, AnnotationCore> curMatchedAnnotations;
		KappaCompareCombi curKappaCC;
		
		for (CompareCombi cc : compareSegments) {
			curMatchedAnnotations = compareUtil.matchAnnotations(cc, minimalOverlapPercentage);
			curKappaCC = calculateAgreementTables(cc, curMatchedAnnotations);
				
			modCompareCombis.add(curKappaCC);
			
			curProgress += perCombi;
			progressUpdate((int) curProgress, null);
		}
		
		progressUpdate((int) curProgress, "Starting global agreement calculations...");
		
		// collect all labels/codings (per tier set?) 
		SortedSet<String> allValuesGlobal = new TreeSet<String>();
		
		for (KappaCompareCombi kcc : modCompareCombis) {
			allValuesGlobal.addAll(kcc.getAllValues());// includes UNMATCHED
		}
		List<String> globalValuesList = new ArrayList<String>(allValuesGlobal);
		if (globalValuesList.remove(UNMATCHED)) {
			globalValuesList.add(UNMATCHED);// move to the end of the list
		}
		
		AgreementTable globalTable = new AgreementTable("Multi file global", globalValuesList);
		Map<String, TwoSquareTable> globalPerValueTables = new HashMap<String, TwoSquareTable>();
		for (int i = 0; i < globalValuesList.size() - 1; i++) {// don't create a 2x2 table for unmatched
			String code = globalValuesList.get(i);
			globalPerValueTables.put(code, new TwoSquareTable(code));
		}
		
		// re-iterate the kappa combis
		for (KappaCompareCombi kcc : modCompareCombis) {
			mergeGlobalAgreementTables(globalTable, kcc.getOverallTable());
			
			mergePerValueTables(globalPerValueTables, kcc.getPerValueTables()); 
		}
		// hier... calculate kappa agreement, raw agreement, ..., ipf kappa?
		CKappa ckappa = new CKappa();
		globalCombi = new KappaCompareCombi(null, null);
		globalCombi.setAllValues(globalValuesList);
		globalCombi.setOverallTable(globalTable);
		globalCombi.setPerValueTables(globalPerValueTables);
		
		Iterator<String> perValueIter = globalPerValueTables.keySet().iterator();
		String keyIter;
		TwoSquareTable iterTable;
		while (perValueIter.hasNext()) {
			keyIter = perValueIter.next();
			iterTable = globalPerValueTables.get(keyIter);

			globalCombi.getPerValueAgreement().put(keyIter, ckappa.calcKappa(iterTable.getTable()));

		}
		// hier... calculate overall agreement values as well?
		if (perFileOutput) {
			// hier...
		}
		// hier...
		progressComplete(String.format("Completed calculations of %d pairs of tiers.", combiCount));
	}


	@Override
	public void writeResultsAsText(File toFile, String encoding)
			throws IOException {
		if (globalCombi == null || globalCombi.getAllValues().isEmpty()){
			throw new IOException("There are no results to save.");// throw other type of exception
		}
		if (toFile == null) {
			throw new IOException("There is no file location specified.");
		}
		BufferedWriter writer = null;

        try {
            FileOutputStream out = new FileOutputStream(toFile);
            OutputStreamWriter osw = null;

            try {
                osw = new OutputStreamWriter(out, encoding);
            } catch (UnsupportedCharsetException uce) {
                osw = new OutputStreamWriter(out, "UTF-8");
            }

            writer = new BufferedWriter(osw);
            final String NL = "\n";
            final String TAB = "\t";
            // write "header" date and time
            writer.write(String.format("Output created: %tD %<tT",  Calendar.getInstance()));
            writer.write(NL); // or writer.newLine();//??
            writer.write("Overall kappa: " + globalCombi.getOverallAgreement());//??
            writer.write(NL);
            writer.write("Kappa per value:");
            writer.write(NL);
            List<String> allValues = globalCombi.getAllValues();
            for (String value : allValues) {
            	writer.write(value);
            	writer.write(TAB);
            	writer.write(String.valueOf(globalCombi.getAgreementForValue(value)));
            	writer.write(NL);
            }
            // write the agreement table 
            writer.write("Overall agreement table:");
            writer.write(NL);
            AgreementTable overallTable = globalCombi.getOverallTable();
            List<String> labels = overallTable.getLabels();
            
            for (int i = 0; i < labels.size(); i++) {
            	writer.write(labels.get(i));
            	writer.write(TAB);
            	//int[] row = overallTable.getTable()[i];
            	for (int j = 0; j < overallTable.getTable()[i].length; j++) {
            		writer.write(String.valueOf(overallTable.getTable()[i][j]));
            		writer.write(TAB);
            	}
            	writer.write(NL);
            }
            writer.flush();
        } catch (Throwable tr) {
            // FileNotFound, Security or UnsupportedEncoding exceptions
            throw new IOException("Cannot write to file: " + tr.getMessage());
        } finally {
        	if (writer != null) {
	        	try {
	        		writer.close();
	        	} catch (Throwable t){
	        		
	        	}
        	}
        }
	}

	private KappaCompareCombi calculateAgreementTables(CompareCombi cc, 
			Map<AnnotationCore, AnnotationCore> matchedAnnotations) {
		KappaCompareCombi curKappaCC = new KappaCompareCombi(cc.getFirstUnit(), cc.getSecondUnit());
		
	// step 1: extract all values that are in the two lists of annotations
		SortedSet<String> allValues = new TreeSet<String>();
		
		List<AnnotationCore> ac1List = cc.getFirstUnit().annotations;
		for (AnnotationCore ac : ac1List) {
			allValues.add(ac.getValue());
		}
		List<AnnotationCore> ac2List = cc.getSecondUnit().annotations;
		for (AnnotationCore ac : ac2List) {
			allValues.add(ac.getValue());
		}
		List<String> allValuesList = new ArrayList<String>(allValues);
		allValuesList.add(UNMATCHED);
		curKappaCC.setAllValues(allValuesList);
		
	// step 2: process the matched annotations, add them to the proper agreement matrices
		// handle situation where there are no values? or will there always be the empty string?
		AgreementTable globalTable = new AgreementTable("File global", allValuesList);
		Map<String, TwoSquareTable> perValueTable = new HashMap<String, TwoSquareTable>();
		for (int i = 0; i < allValuesList.size() - 1; i++) {// don't create a 2x2 table for unmatched
			String code = allValuesList.get(i);
			perValueTable.put(code, new TwoSquareTable(code));
		}
		
		AnnotationCore ac1, ac2;
		String s1, s2;
		
		if (matchedAnnotations != null) {
			Iterator<AnnotationCore> acIter = matchedAnnotations.keySet().iterator();
			
			while (acIter.hasNext()) {
				ac1 = acIter.next();
				ac2 = matchedAnnotations.get(ac1);
				s1 = ac1.getValue();
				s2 = ac2.getValue();

				// update tables
				if (s1.equals(s2)) {
					globalTable.increment(s1);
					perValueTable.get(s1).increment(0, 0);// update first cell, both raters said Yes/1 here
					// all other 2x2 value tables need to increment cell at [1, 1]
				} else {
					globalTable.increment(s1, s2);// -> Note: this way R1 along the vertical axis, R2 along horizontal
					perValueTable.get(s1).increment(0, 1);// -> Note: R1 along vertical axis
					perValueTable.get(s2).increment(1, 0);// -> Note: R2 along horizontal axis
					// all other 2x2 value tables need to increment cell at [1, 1]
				}
			}
		}
		
	// step 3: process unmatched annotations from the first set of annotations, the first "unit"
		for (AnnotationCore aco : ac1List) {
			if (!matchedAnnotations.containsKey(aco)) {// not processed yet
				s1 = aco.getValue();
				globalTable.increment(s1, UNMATCHED);// ->
				perValueTable.get(s1).increment(0, 1);//->
			}
		}
		
	// step 4: process unmatched annotations from the second set of annotations, the second "unit"
		for (AnnotationCore aco : ac2List) {
			if (!matchedAnnotations.containsValue(aco)) {// not processed yet
				s2 = aco.getValue();
				globalTable.increment(UNMATCHED, s2);// ->
				perValueTable.get(s2).increment(1, 0);//->
			}
		}
		
	// step 5: set the value of the [1, 1] cell of the per value tables,
		// based on totals and the value in other 3 cells
		int numMatches = matchedAnnotations.size();
		int totalCount = numMatches + (ac1List.size() - numMatches) + (ac2List.size() - numMatches);
		
		Iterator<String> labelIter = perValueTable.keySet().iterator();
		while (labelIter.hasNext()) {
			TwoSquareTable tst = perValueTable.get(labelIter.next());
			int[][] curTable = tst.getTable();
			int curTotal = curTable[0][0] + curTable[0][1] + curTable[1][0];
			curTable[1][1] = totalCount - curTotal;
		}
		
	// step 6 add everything to the combi
		curKappaCC.setOverallTable(globalTable);
		curKappaCC.setPerValueTables(perValueTable);
		
		return curKappaCC;
	}
	
	/**
	 * Adds contents from the second table to the first.
	 * 
	 * @param intoTable receiver, not null
	 * @param fromTable supplier, not null
	 */
	private void mergeGlobalAgreementTables(AgreementTable intoTable, AgreementTable fromTable) {
		int numRows = fromTable.getLabels().size();
		
		// nested loops
		for (int i = 0; i < numRows; i++) {
			for (int j = 0; j < numRows; j++) {// columns
				intoTable.addCount(fromTable.getLabels().get(i), fromTable.getLabels().get(j), 
						fromTable.getTable()[i][j]);
			}
		}
	}
	
	/**
	 * Merges 2x2 value tables with the global table for that value.
	 * 
	 * @param intoMap the global map, the global tables, not null
	 * @param fromMap the map of a single comparison set, not null  
	 */
	private void mergePerValueTables(Map<String, TwoSquareTable> intoMap, Map<String, TwoSquareTable> fromMap) {
		TwoSquareTable intoTable;
		TwoSquareTable fromTable;
		String codeValue;
		Iterator<String> codeIter = fromMap.keySet().iterator();
		
		while (codeIter.hasNext()) {
			codeValue = codeIter.next();
			intoTable = intoMap.get(codeValue);
			fromTable = fromMap.get(codeValue);
			
			intoTable.getTable()[0][0] += fromTable.getTable()[0][0];
			intoTable.getTable()[1][0] += fromTable.getTable()[1][0];
			intoTable.getTable()[0][1] += fromTable.getTable()[0][1];
			intoTable.getTable()[1][1] += fromTable.getTable()[1][1];
		}
	}
}
