package mpi.eudico.client.annotator.interlinear.edit;

import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Window;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;

import javax.swing.JLabel;
import javax.swing.JScrollPane;
import javax.swing.JWindow;
import javax.swing.UIManager;

import mpi.eudico.client.annotator.interlinear.edit.event.SuggestionSelectionEvent;
import mpi.eudico.client.annotator.interlinear.edit.event.SuggestionSelectionListener;

/**
 * A window class that can show suggestions in a kind of popup near the original annotation.
 * 
 * @author Han Sloetjes
 */
public class SuggestionWindow extends JWindow implements MouseListener, 
MouseMotionListener, ComponentListener {
	private JScrollPane suggestionScrollPane; 
	private SuggestionSetSelector selector;
	private SuggestionSelectionListener suggestionSelectionListener;
	private JLabel closeLabel;
	
    private Point dragStartPoint;
    private Point dragEndPoint;
    private boolean dragCorner = false;
    private Insets windowMargins;
    private int scrollBarWidth;
	
	/**
	 * Constructor with a frame as the owner.
	 * 
	 * @param frame the owner
	 * @param selector the suggestion set selector component
	 */
	public SuggestionWindow(Frame frame, SuggestionSetSelector selector) {
		super(frame);
		this.selector = selector;
		initComponents();
	}

	/**
	 * Constructor with a window as the owner.
	 * 
	 * @param window the owner
	 * @param selector the suggestion set selector component
	 */
	public SuggestionWindow(Window window, SuggestionSetSelector selector) {
		super(window);
		this.selector = selector;
		initComponents();
	}
	
	/**
	 * Initializes the ui components, adds the listeners.
	 */
	private void initComponents() {
		suggestionScrollPane = new JScrollPane(selector);
		getContentPane().setLayout(null);
		getContentPane().add(suggestionScrollPane);
		closeLabel = new JLabel("Close");
		closeLabel.setFont(closeLabel.getFont().deriveFont(8f));
		getContentPane().add(closeLabel);
		windowMargins = new Insets(12, 10, 10, 10);
		
		Object sbWidth = UIManager.getDefaults().get("ScrollBar.width");
        if (sbWidth != null) {
            scrollBarWidth = ((Integer) sbWidth).intValue();
        }
		suggestionScrollPane.addMouseListener(this);
		suggestionScrollPane.addMouseMotionListener(this);
		closeLabel.addMouseListener(this);
		
		addMouseListener(this);
		addMouseMotionListener(this);
		addComponentListener(this);
		positionComponents();
	}
	
	/**
	 * Repositions the components of this window.
	 */
	private void positionComponents() {
		int w = getWidth();
		int h = getHeight();
		
		Dimension closeDim = closeLabel.getPreferredSize();
		int cx = w - closeDim.width - windowMargins.right;
		if (cx > windowMargins.left) {
			if (!closeLabel.isVisible()) {
				closeLabel.setVisible(true);
			}
			closeLabel.setBounds(cx, 0, closeDim.width, closeDim.height);
		} else {
			closeLabel.setVisible(false);
		}
		if (w > windowMargins.left + windowMargins.right && h > windowMargins.top + windowMargins.bottom) {
			if (!suggestionScrollPane.isVisible()) {
				suggestionScrollPane.setVisible(true);
			}
			suggestionScrollPane.setBounds(windowMargins.left, windowMargins.top, w - windowMargins.left - windowMargins.right, 
					h - windowMargins.top - windowMargins.bottom);
		} else {
			suggestionScrollPane.setVisible(false);
		}
	}

	/**
	 * Removes the listeners from this window.
	 */
	private void removeListeners() {
		removeMouseListener(this);
		removeMouseMotionListener(this);
		suggestionScrollPane.removeMouseListener(this);
		suggestionScrollPane.removeMouseMotionListener(this);
		closeLabel.removeMouseListener(this);
		suggestionSelectionListener = null;
	}
	
	/**
	 * Removes listeners and closes the window.
	 */
	private void closeSelectionWindow() {
		removeListeners();
		setVisible(false);
		dispose();
	}
	
	/**
	 * Sets the listener to be notified of a suggestion selection event.
	 * Currently there can only be one such listener.
	 * 
	 * @param listener the new suggestion selection listener
	 */
	public void setSuggestionSelectionListener(SuggestionSelectionListener listener) {
		suggestionSelectionListener = listener;
	}

	// ######## mouse motion listener
	@Override
	public void mouseDragged(MouseEvent e) {
		dragEndPoint = e.getPoint();
		int dx = dragEndPoint.x - dragStartPoint.x;
		int dy = dragEndPoint.y - dragStartPoint.y;
		if (!dragCorner) {
			this.setLocation(getLocation().x + dx, getLocation().y + dy);
		} else {
			Dimension size = getSize();
			setSize(size.width + dx, size.height + dy);
			dragStartPoint = dragEndPoint;
			//positionComponents();
		}
		
	}

	@Override
	public void mouseMoved(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	// ####### mouse listener
	@Override
	public void mouseClicked(MouseEvent e) {
		if (e.getSource() == closeLabel) {
			closeSelectionWindow();
		} else if (e.getSource() == suggestionScrollPane) {
//			System.out.println("Scrollpane " + e.getPoint());
			int xp = e.getPoint().x + suggestionScrollPane.getHorizontalScrollBar().getValue();
			int xy = e.getPoint().y + suggestionScrollPane.getVerticalScrollBar().getValue();
			
			int selectedIndex = selector.getSuggestionIndexAtPoint(xp, xy);
			//System.out.println("Selected index: " + selectedIndex);
			if (selectedIndex > -1 && suggestionSelectionListener != null) {
				SuggestionSelectionEvent event = new SuggestionSelectionEvent(selector.getModel(), selectedIndex);
				suggestionSelectionListener.suggestionSelected(event);
			}
			closeSelectionWindow();
		}
		
	}

	@Override
	public void mouseEntered(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void mouseExited(MouseEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void mousePressed(MouseEvent e) {
		dragStartPoint = e.getPoint();
		if (getWidth() - dragStartPoint.x < 10 && getHeight() - dragStartPoint.y < 10) {
			dragCorner = true;
		}
	}

	@Override
	public void mouseReleased(MouseEvent e) {
		dragCorner = false;
	}

	@Override
	public void componentHidden(ComponentEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void componentMoved(ComponentEvent e) {
		// TODO Auto-generated method stub
		
	}

	@Override
	public void componentResized(ComponentEvent e) {
		positionComponents();
	}

	@Override
	public void componentShown(ComponentEvent e) {
		positionComponents();
		selector.init();
		suggestionScrollPane.revalidate();
		this.setSize(selector.getPreferredSize().width + windowMargins.left + windowMargins.right + scrollBarWidth, 
				selector.getPreferredSize().height + windowMargins.top + windowMargins.bottom + scrollBarWidth);
	}


}
