package mpi.eudico.client.annotator.tier;

import java.util.Set;
import java.util.Vector;

import javax.swing.table.AbstractTableModel;

/**
 * Table model to display a table with checkable items
 * @author Jeffrey Lemein
 * @author Han Sloetjes
 * @version March, 2011
 * @version July, 2014 Removed reference to the JTable, the data model should preferably be 
 * unaware of a view object; views will receive events if they are registered as listener.  
 */
@SuppressWarnings("serial")
public class SelectableContentTableModel extends AbstractTableModel {
	private String[] columnNames;
	private Object[][] rowData;
	
	public SelectableContentTableModel( Set<? extends Object> content){
		columnNames = new String[2];
		rowData = new Object[content.size()][2];
		
		Object[] contentArray = content.toArray();
		
		for( int i =0; i<contentArray.length; i++ ){
			rowData[i][0] = Boolean.FALSE;
			rowData[i][1] = contentArray[i];
		}
	}
	
	/**
	 * Creates a selectable content table model in which the data for two columns can be specified.
	 * Both arrays of data should be of the same length
	 * @param column1Data data for first column
	 * @param column2Data data for second column
	 */
	public SelectableContentTableModel( Object[]... columnData ){
		int nrArguments = columnData.length;
		columnNames = new String[nrArguments+1];
		
		rowData = new Object[columnData[0].length][nrArguments+1];
		
		for( int i =0; i<columnData[0].length; i++ ){
			rowData[i][0] = Boolean.FALSE;
			for( int j=0; j<nrArguments; j++ )
				rowData[i][j+1] = columnData[j][i];
		}
	}
	
	public int getRowCount(){
		return rowData.length;			
	}
	
	public int getColumnCount(){
		return columnNames.length;
	}
	
	public Object getValueAt(int row, int column){
		return rowData[row][column];
	}
	
	public void setValueAt( Object object, int row, int column ){
		if( getColumnCount() > column && getRowCount() > row ){
			rowData[row][column] = object;
			fireTableRowsUpdated(row, row);
		}
	}
	
	public void selectAll(){
		if( getColumnCount() > 0 )
			for( int i = 0; i < getRowCount(); i++ )
				setValueAt( true, i, 0 );
		fireTableDataChanged();
	}
	
	public Vector<Object> getSelectedValues(){
		Vector<Object> selectedValues = new Vector<Object>();
		
		for( int i=0; i<rowData.length; i++ )
			if( (Boolean) rowData[i][0] )
				selectedValues.add(rowData[i][1]);
		
		return selectedValues;
	}
	
	public void selectNone(){
		if( getColumnCount() > 0 )
			for( int i = 0; i < getRowCount(); i++ )
				setValueAt( false, i, 0 );
		fireTableDataChanged();
	}	
	
    public void moveUp(int row) { 
    	// can throw ArrayIndexOutOfBoundsException here 
    	if(row > 0 && row < rowData.length){
    		Object val1 = rowData[row-1][0];
    		Object val2 = rowData[row-1][1];
    		
    		Object obj1 = rowData[row][0];
    		Object obj2 = rowData[row][1];
    		
    		
    		rowData[row-1][0] = obj1;
    		rowData[row-1][1] = obj2;
    		
    		rowData[row][0] = val1;
    	    rowData[row][1] = val2;
    	    
    	    fireTableRowsUpdated(row - 1, row);
//    	    fireTableDataChanged();
    	} 
    }
    
    public void moveDown(int row) { 
    	// can throw ArrayIndexOutOfBoundsException here
    	if(row < rowData.length - 1 && row >= 0){
    		Object val1 = rowData[row+1][0];
    		Object val2 = rowData[row+1][1];
    		
    		Object obj1 = rowData[row][0];
    		Object obj2 = rowData[row][1];
    		
    		
    		rowData[row+1][0] = obj1;
    		rowData[row+1][1] = obj2;
    		
    		rowData[row][0] = val1;
    	    rowData[row][1] = val2;
    	    
    	    fireTableRowsUpdated(row, row + 1);
//    	    fireTableDataChanged();
    	} 
    }
	
	public boolean nothingSelected(){
		for( int r=0; r<getRowCount(); r++ )
			if( (Boolean)getValueAt(r,0) == true )
				return false;
		
		return true;
	}
	
	public boolean isCellEditable(int row, int col) {
        //only the first column is editable (because it contains the checkboxes)
		return col <= 0;
    }
	
	public Class<? extends Object> getColumnClass(int c) {
		// can throw ArrayIndexOutOfBoundsException here
		return getValueAt(0, c).getClass();
    }
}
