package mpi.eudico.server.corpora.clomimpl.dobes;

import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;

import mpi.eudico.client.annotator.prefs.RecentLanguages;
import mpi.eudico.server.corpora.clom.Annotation;
import mpi.eudico.server.corpora.clom.AnnotationDocEncoder;
import mpi.eudico.server.corpora.clom.EncoderInfo;
import mpi.eudico.server.corpora.clom.ExternalReference;
import mpi.eudico.server.corpora.clom.Property;
import mpi.eudico.server.corpora.clom.Tier;
import mpi.eudico.server.corpora.clom.TimeOrder;
import mpi.eudico.server.corpora.clom.TimeSlot;
import mpi.eudico.server.corpora.clom.Transcription;
import mpi.eudico.server.corpora.clomimpl.abstr.AbstractAnnotation;
import mpi.eudico.server.corpora.clomimpl.abstr.AlignableAnnotation;
import mpi.eudico.server.corpora.clomimpl.abstr.ExternalReferenceImpl;
import mpi.eudico.server.corpora.clomimpl.abstr.LicenseRecord;
import mpi.eudico.server.corpora.clomimpl.abstr.LinkedFileDescriptor;
import mpi.eudico.server.corpora.clomimpl.abstr.MediaDescriptor;
import mpi.eudico.server.corpora.clomimpl.abstr.PropertyImpl;
import mpi.eudico.server.corpora.clomimpl.abstr.RefAnnotation;
import mpi.eudico.server.corpora.clomimpl.abstr.SVGAlignableAnnotation;
import mpi.eudico.server.corpora.clomimpl.abstr.TierImpl;
import mpi.eudico.server.corpora.clomimpl.abstr.TranscriptionImpl;
import mpi.eudico.server.corpora.clomimpl.type.Constraint;
import mpi.eudico.server.corpora.clomimpl.type.LinguisticType;
import mpi.eudico.server.corpora.lexicon.LexiconLink;
import mpi.eudico.server.corpora.lexicon.LexiconQueryBundle2;
import mpi.eudico.server.corpora.util.ServerLogger;
import mpi.eudico.util.CVEntry;
import mpi.eudico.util.ControlledVocabulary;
import mpi.eudico.util.ExternalCV;
import mpi.eudico.util.IoUtil;
import mpi.eudico.util.multilangcv.LangInfo;
import mpi.eudico.util.multilangcv.LanguageCollection;

import org.w3c.dom.Element;

/**
 * Encodes a Transcription to EAF 2.8 format and saves it.
 *
 * @version Aug 2005 Identity removed
 * @version Feb 2006 Constraint Included In added as well as LinkedFileDescriptors
 * @version Dec 2006 attribute Annotator of Tier added and writing of document 
 * level Properties.
 * @version Nov 2007 added support for attribute RELATIVE_MEDIA_URL of MEDIA_DESCRIPTOR
 *  and RELATIVE_LINK_URL of LINKED_FILE_DESCRIPTOR
 * @version May 2008 added attributes and elements concerning DCR references
 * @version Dec 2010 added support for external Controlled Vocabularies and links to 
 * a Lexicon (or Lexicon services)  
 */
public class EAF28Encoder implements AnnotationDocEncoder, ServerLogger {
    /** the version string of the format / dtd */
    public static final String VERSION = "2.8";
    public static boolean debug = false;
    /**
     * Creates a DOM and saves.
     *
     * @param theTranscription the Transcription to store
     * @param encoderInfo additional information for encoding
     * @param tierOrder preferred tier ordering; should be removed
     * @param path the output path
     */
    @Override
	public void encodeAndSave(Transcription theTranscription,
        EncoderInfo encoderInfo, List<TierImpl> tierOrder, String path)
    	   throws IOException{
    	Element documentElement = createDOM(theTranscription, tierOrder, path);
        save(documentElement, path);
    }

    /**
     * Saves a template eaf of the Transcription; everything is saved except for
     * the annotations
     *
     * @param theTranscription the Transcription to store
     * @param tierOrder preferred tier ordering; should be removed
     * @param path the output path
     */
    public void encodeAsTemplateAndSave(Transcription theTranscription,
        List<TierImpl> tierOrder, String path) throws IOException{
        Element documentElement = createTemplateDOM(theTranscription, tierOrder, path);
        save(documentElement, path);
    }

    /**
     * Create the DOM tree and returns the document element.
     *
     * @param theTranscription the Transcription to save (not null)
     * @param tierOrder the preferred ordering of the tiers
     *
     * @return the document element
     */
    public static Element createDOM(Transcription theTranscription, List<TierImpl> tierOrder, String path) {
    	long beginTime = System.currentTimeMillis();
    	if (debug) {
    		System.out.println("Encoder creating DOM...");
    	}
        Map<String, Element> tierElements = new HashMap<String, Element>(); // for temporary storage of created tier Elements
        Map<TimeSlot, String> timeSlotIds = new HashMap<TimeSlot, String>(); // for temporary storage of generated tsIds
        Map<Annotation, String> annotationIds = new HashMap<Annotation, String>(); // for temporary storage of generated annIds
        //Map svgIds = new HashMap(); // for temporary storage of generated svg ids.
        // may 2008 create and store ids and elements for external reference objects. 
        // if different references to the same external object are found it is stored only once
        GetExtRefIdParams getExtRefIdParams = new GetExtRefIdParams();
        
        // For saving Lexicon Query Bundles:
        HashMap<String, LexiconQueryBundle2> lexRefs = new HashMap<String, LexiconQueryBundle2>();
        int lexRefIndex = 1;
        
        List<Locale> usedLocales = new ArrayList<Locale>(); // for storage of used locales

        TranscriptionImpl attisTr = (TranscriptionImpl) theTranscription;

        if (attisTr == null) {
            LOG.warning(
                "[[ASSERTION FAILED]] TranscriptionStore/storeTranscription: theTranscription is null");
        }

        EAF28 eafFactory = null;

        try {
            eafFactory = new EAF28();
        } catch (Exception ex) {
            ex.printStackTrace();
        }

        // ANNOTATION_DOCUMENT
        String dateString = getDate();

        String author = attisTr.getAuthor();

        if (author == null) {
            author = "unspecified";
        }

        Element annotDocument = eafFactory.newAnnotationDocument(dateString,
                author, VERSION);
        eafFactory.appendChild(annotDocument);
        
        // LICENSEs - new in 2.8
        List<LicenseRecord> licenses = theTranscription.getLicenses();
        if (licenses != null) {
            for (LicenseRecord lr : licenses) {
                Element license = eafFactory.newLicense(lr);
                annotDocument.appendChild(license);
            }
        }

        // HEADER
        Element header = eafFactory.newHeader(""); // mediaFile maintained for compat with 1.4.1
        annotDocument.appendChild(header);

        Iterator<MediaDescriptor> mdIter = attisTr.getMediaDescriptors().iterator();

        while (mdIter.hasNext()) {
            MediaDescriptor md = mdIter.next();

            String origin = null;

            if (md.timeOrigin != 0) {
                origin = String.valueOf(md.timeOrigin);
            }

            String extrFrom = null;

            if ((md.extractedFrom != null) && (md.extractedFrom != "")) {
                extrFrom = md.extractedFrom;
            }

            Element mdElement = eafFactory.newMediaDescriptor(md.mediaURL,
                    md.relativeMediaURL, md.mimeType, origin, extrFrom);

            header.appendChild(mdElement);
        }

        Iterator<LinkedFileDescriptor> lfdIt = attisTr.getLinkedFileDescriptors().iterator();
        LinkedFileDescriptor lfd;

        while (lfdIt.hasNext()) {
            lfd = lfdIt.next();

            String origin = null;

            if (lfd.timeOrigin != 0) {
                origin = String.valueOf(lfd.timeOrigin);
            }

            Element lfdElement = eafFactory.newLinkedFileDescriptor(lfd.linkURL,
                    lfd.relativeLinkURL, lfd.mimeType, origin, lfd.associatedWith);
            header.appendChild(lfdElement);
        }
        // jan 2007 add document properties
        int lastUsedAnnId = 0;
        ArrayList<Property> props = attisTr.getDocProperties();
        if (props.size() > 0) {
        	PropertyImpl prop;
        	for (int i = 0; i < props.size(); i++) {
        		prop = (PropertyImpl) props.get(i);
        		if (prop != null && 
        				(prop.getName() != null || prop.getValue() != null)) {
        			// special treatment of last used ann id
        			if (prop.getName().equals("lastUsedAnnotationId")) {
        				if (prop.getValue() != null) {
	        				try {
	        					lastUsedAnnId = Integer.parseInt((String) prop.getValue());
	        				} catch (NumberFormatException nfe) {
	        					System.out.println("Could not retrieve the last used annotation id.");
	        				}
        				}
        				continue;
        			}
        			
        			if (prop.getValue() != null) {
        				header.appendChild(eafFactory.newProperty(
            					prop.getName(), prop.getValue().toString())); 				
        			} else {
        				header.appendChild(eafFactory.newProperty(
            					prop.getName(), null));
        			}
        			
        		}
        	}
        }
        
        if (debug) {
        	System.out.println("Header creation took: " + (System.currentTimeMillis() - beginTime) + " ms");
        	beginTime = System.currentTimeMillis();
        }
        // TIME_ORDER
        TimeOrder timeOrder = attisTr.getTimeOrder();

        // HB, July 19, 2001: cleanup unused TimeSlots first
        timeOrder.pruneTimeSlots();

        Element timeOrderElement = eafFactory.newTimeOrder();
        annotDocument.appendChild(timeOrderElement);

        int index = 1;

        Enumeration<TimeSlot> tsElements = timeOrder.elements();

        while (tsElements.hasMoreElements()) {
            TimeSlot ts = tsElements.nextElement();

            Element tsElement = null;
            String tsId = "ts" + index;

            // store ts with it's id temporarily
            timeSlotIds.put(ts, tsId);

            if (ts.getTime() != TimeSlot.TIME_UNALIGNED) {
                tsElement = eafFactory.newTimeSlot(tsId, ts.getTime());
            } else {
                tsElement = eafFactory.newTimeSlot(tsId);
            }

            timeOrderElement.appendChild(tsElement);

            index++;
        }
        if (debug) {
        	System.out.println("TimeSlots creation took: " + (System.currentTimeMillis() - beginTime) + " ms");
        	beginTime = System.currentTimeMillis();
        }
        // TIERS
        //Vector tiers = attisTr.getTiers();
        // HS Nov 2009 always store tiers in the order they are in the transcription, for new files this means
        // creation order
        List<Tier> storeOrder = new ArrayList<Tier>(attisTr.getTiers());
        /*
        Vector storeOrder = new Vector(tierOrder); // start with tiers in specified order
        Iterator tIter = tiers.iterator();

        while (tIter.hasNext()) { // add other tiers in document order

            Tier t = (Tier) tIter.next();

            if (!storeOrder.contains(t)) {
                storeOrder.add(t);
            }
        }
		*/
        int svgIndex = 1; // used to create svg id values
        //int annIndex = 0;
        int annIndex = lastUsedAnnId;

        Iterator<Tier> tierIter = storeOrder.iterator();

        while (tierIter.hasNext()) {
            TierImpl t = (TierImpl) tierIter.next();

            String id = t.getName();
            String participant = (String) t.getMetadataValue("PARTICIPANT");
            String annotator = (String) t.getMetadataValue("ANNOTATOR");
            String lingType = t.getLinguisticType().getLinguisticTypeName();

            if (lingType == null) {
                lingType = "not specified";
            }

            //Locale lang = (Locale) t.getMetadataValue("DEFAULT_LOCALE");
            Locale lang = t.getDefaultLocale();
            // HS Feb 2010 allow null for Locale, use empty string internally
//            if (lang == null) {
//                lang = new Locale("not specified", "", "");
//            }

            // check is quick solution, TreeSet would do this but compareTo causes ClassCastException
            if (lang != null && !usedLocales.contains(lang)) {
                usedLocales.add(lang);
            }

            String parentName = null;

            if (t.getParentTier() != null) {
                parentName = t.getParentTier().getName();
            }
            
            String extRef = t.getExtRef();		// new in 2.8
            String langRef = t.getLangRef();	// new in 2.8

            Element tierElement = eafFactory.newTier(id, participant, annotator, 
            		lingType, lang, parentName, extRef, langRef);
            annotDocument.appendChild(tierElement);

            tierElements.put(t.getName(), tierElement); // store for later use

            List<Annotation> annotations = t.getAnnotations();

            for (Annotation ann : annotations) {
                //annotation has already an id
                if ((ann.getId() != null) && !ann.getId().equals("")) {
                    annotationIds.put(ann, ann.getId());
                } else {
                    //create an id that isn't yet in the transcription
                	   annIndex++;
                    annotationIds.put(ann, "a" + annIndex);
                }
            }

        }
        header.appendChild(eafFactory.newProperty("lastUsedAnnotationId", String.valueOf(annIndex)));

        // ANNOTATIONS
        // second pass. Actually creates and adds Annotation Elements
        Iterator<Tier> tierIter2 = storeOrder.iterator();

        while (tierIter2.hasNext()) {
            TierImpl t = (TierImpl) tierIter2.next();

            List<Annotation> annotations = t.getAnnotations();

            Iterator<Annotation> annotIter2 = annotations.iterator();

            while (annotIter2.hasNext()) {
                Annotation ann = annotIter2.next();
                // may 2008 store and add external reference id refs
                String extRefId = null;
                
				if (ann instanceof AbstractAnnotation) {
					final List<ExternalReference> extRefs = ((AbstractAnnotation) ann).getExtRefs();
					
					if (extRefs != null) {
	                	for (ExternalReference thisExtRef : extRefs) {
	                		String tmpExtRefId = getExtRefId(getExtRefIdParams, thisExtRef);
	                		if (extRefId != null && !extRefId.isEmpty()) {
	                			extRefId += " " + tmpExtRefId;
	                		} else {
	                			extRefId = tmpExtRefId;
	                		}
						}
	                }
				}
                
                Element annElement = eafFactory.newAnnotation();
                tierElements.get(t.getName()).appendChild(annElement);

                Element annSubElement = null;

                String annId = annotationIds.get(ann);

                if (ann instanceof AlignableAnnotation) {
                    String beginTsId = timeSlotIds.get(((AlignableAnnotation) ann).getBegin());
                    String endTsId = timeSlotIds.get(((AlignableAnnotation) ann).getEnd());
                  

                    if (ann instanceof SVGAlignableAnnotation) {
                        if (((SVGAlignableAnnotation) ann).getShape() != null) {
                            String svgId = "ga" + svgIndex;
                            ((SVGAlignableAnnotation) ann).setSVGElementID(svgId);
                            svgIndex++;
                            //svgIds.put(ann, svgId);
                            annSubElement = eafFactory.newAlignableAnnotation(annId,
                                    beginTsId, endTsId, svgId, extRefId, ann.getCVEntryId());
                        } else {
                            ((SVGAlignableAnnotation) ann).setSVGElementID(null);
                            annSubElement = eafFactory.newAlignableAnnotation(annId,
                                    beginTsId, endTsId, null, extRefId, ann.getCVEntryId());
                        }
                    } else {
                        annSubElement = eafFactory.newAlignableAnnotation(annId,
                                beginTsId, endTsId, null, extRefId, ann.getCVEntryId());
                    }
                } else if (ann instanceof RefAnnotation) {
                    String refId = null;
                    String prevId = null;
                    List<Annotation> refs = ((RefAnnotation) ann).getReferences();
                    RefAnnotation prev = ((RefAnnotation) ann).getPrevious();

                    // for the moment, take the first, if it exists
                    if (refs.size() > 0) {
                        refId = annotationIds.get(refs.get(0));
                    }

                    if (prev != null) {
                        prevId = annotationIds.get(prev);
                    }

                    annSubElement = eafFactory.newRefAnnotation(annId, refId,
                            prevId, extRefId, ann.getCVEntryId());
                }

                annElement.appendChild(annSubElement);

                // ANNOTATION_VALUE
                Element valueElement;
        		valueElement = eafFactory.newAnnotationValue(ann.getValue());

        		annSubElement.appendChild(valueElement);
            }
        }
        if (debug) {
        	System.out.println("Tiers and Annotations creation took: " + (System.currentTimeMillis() - beginTime) + " ms");
        	beginTime = System.currentTimeMillis();
        }
        // LINGUISTIC_TYPES
        List<LinguisticType> lTypes = attisTr.getLinguisticTypes();

        if (lTypes != null) {
            Iterator<LinguisticType> typeIter = lTypes.iterator();

            while (typeIter.hasNext()) {
                // HB, april 24, 2002: for the moment, just store lt name
                LinguisticType lt = typeIter.next();
                // may 2008 store and add external reference id ref
                String extRefId = null;
                if (lt.getDataCategory() != null && lt.getDataCategory().length() > 0) {
                	ExternalReferenceImpl eri = new ExternalReferenceImpl(lt.getDataCategory(), ExternalReference.ISO12620_DC_ID);
                	extRefId = getExtRefId(getExtRefIdParams, eri);
                }
                
                String stereotype = null;

                if (lt.hasConstraints()) {
                    stereotype = Constraint.stereoTypes[lt.getConstraints()
                                                          .getStereoType()];
                    stereotype = stereotype.replace(' ', '_');
                }

                // Create a LEXICON_REF if there is a Lexicon Query Bundle used.
                String lexRef = null;
                LexiconQueryBundle2 queryBundle = lt.getLexiconQueryBundle();
                if(queryBundle != null) {
                	lexRef = "lr" + lexRefIndex++;
                	lexRefs.put(lexRef, queryBundle);
                }
                
                Element typeElement = eafFactory.newLinguisticType(lt.getLinguisticTypeName(),
                        lt.isTimeAlignable(), lt.hasGraphicReferences(),
                        stereotype, lt.getControlledVocabularyName(), extRefId, 
                        lexRef);
               
                annotDocument.appendChild(typeElement);
            }
        }

        if (debug) {
        	System.out.println("Linguistic Types creation took: " + (System.currentTimeMillis() - beginTime) + " ms");
        	beginTime = System.currentTimeMillis();
        }
        
        // LOCALES
        Iterator<Locale> locIter = usedLocales.iterator();

        while (locIter.hasNext()) {
            Locale l = locIter.next();
            Element locElement = eafFactory.newLocale(l);
            annotDocument.appendChild(locElement);
        }
        
        // <LANGUAGE>s

        List<ControlledVocabulary> conVocs = attisTr.getControlledVocabularies();
        addLanguages(storeOrder, eafFactory, annotDocument, conVocs);

        // HB, 18 jul 02: for the moment manually add relevant Constraints
        // CONSTRAINTS
        Element timeSubdivision = eafFactory.newConstraint(Constraint.stereoTypes[Constraint.TIME_SUBDIVISION].replace(
                    ' ', '_'),
                "Time subdivision of parent annotation's time interval, no time gaps allowed within this interval");
        Element symbSubdivision = eafFactory.newConstraint(Constraint.stereoTypes[Constraint.SYMBOLIC_SUBDIVISION].replace(
                    ' ', '_'),
                "Symbolic subdivision of a parent annotation. Annotations refering to the same parent are ordered");
        Element symbAssociation = eafFactory.newConstraint(Constraint.stereoTypes[Constraint.SYMBOLIC_ASSOCIATION].replace(
                    ' ', '_'), "1-1 association with a parent annotation");
        Element includedIn = eafFactory.newConstraint(Constraint.stereoTypes[Constraint.INCLUDED_IN].replace(
                    ' ', '_'),
                "Time alignable annotations within the parent annotation's time interval, gaps are allowed");

        annotDocument.appendChild(timeSubdivision);
        annotDocument.appendChild(symbSubdivision);
        annotDocument.appendChild(symbAssociation);
        annotDocument.appendChild(includedIn);

        // <CONTROLLED_VOCABULARY>s

        for (ControlledVocabulary cv : conVocs) {
            Element cvElement;

            if (cv instanceof ExternalCV) {
                String extRefId = null;
                
    			// If cv is an external CV, save the cv with a EXT_REF and no entries in the EAF
    			// and save an ECV for caching
    			final ExternalReference externalRef = ((ExternalCV) cv).getExternalRef();
               	extRefId = getExtRefId(getExtRefIdParams, externalRef);
    			cvElement = eafFactory.newControlledVocabulary(cv.getName(),
    					null/*cv.getDescription()*/, extRefId);
    		} else {
    			cvElement = eafFactory.newControlledVocabulary(cv.getName(),
    					null/*cv.getDescription()*/);
    						
    			// <DESCRIPTION>s inside <CONTROLLED_VOCABULARY>
    			int nLangs = cv.getNumberOfLanguages();
    			
    			for (int i = 0; i < nLangs; i++) {
    				String languageId = cv.getLanguageId(i);
    				String description = cv.getDescription(i);
    				Element descriptionElement = eafFactory.newDescription(languageId, description);
    				cvElement.appendChild(descriptionElement);
    			}
    			
    			// <CV_ENTRY_ML>s

    			for (CVEntry entry : cv) {
                    String extRefId = null;

                    final ExternalReference externalRef = entry.getExternalRef();
                	extRefId = getExtRefId(getExtRefIdParams, externalRef);

    				Element entryElement = eafFactory.newCVEntryML(entry.getId(), extRefId);
    				
    				// <CVE_VALUE>s inside <CV_ENTRY_ML>
        			for (int i = 0; i < nLangs; i++) {
        				String languageId = cv.getLanguageId(i);
        				String description = entry.getDescription(i);
        				String value = entry.getValue(i);
        				if (!value.isEmpty() || 
        					(description != null && !description.isEmpty())) {
        					Element valueElement = eafFactory.newCVEntryValue(languageId, value, description);            				
        					entryElement.appendChild(valueElement);
        				}
        			}
        			
        			// Check if there were any children attacked under the CVEntryML
        			// which is required by the schema.
        			if (!entryElement.hasChildNodes()) {
        				// This should not happen, but insert a dummy anyway.
        				String languageId = cv.getLanguageId(0);
    					Element valueElement = eafFactory.newCVEntryValue(languageId, "", "");            				
    					entryElement.appendChild(valueElement);
        			}
    				      				
    				cvElement.appendChild(entryElement);
    			}
    		}


            annotDocument.appendChild(cvElement);
        }

        if (debug) {
        	System.out.println("Constraints and CV's creation took: " + (System.currentTimeMillis() - beginTime) + " ms");
        	beginTime = System.currentTimeMillis();
        }
        
        // LEXICON SERVICES
        // First save the Lexicon Query Bundles
        HashMap<String, LexiconLink> savedLexiconLinks = new HashMap<String, LexiconLink>();
        for(String lexRef : lexRefs.keySet()) {
        	LexiconQueryBundle2 queryBundle = lexRefs.get(lexRef);
        	Element lexSrvcElement = eafFactory.newLexiconReference(lexRef, queryBundle);
        	savedLexiconLinks.put(queryBundle.getLinkName(), queryBundle.getLink());
        	annotDocument.appendChild(lexSrvcElement);
        }
        
        // Then save all Lexicon Links that were not in a Lexicon Query Bundle
        HashMap<String, LexiconLink> lexiconLinks = attisTr.getLexiconLinks();
        for(String linkName : lexiconLinks.keySet()) {
        	if(!savedLexiconLinks.containsKey(linkName)) {
        		Element lexSrvcElement = eafFactory.newLexiconLink("lr" + lexRefIndex++, lexiconLinks.get(linkName));
        		annotDocument.appendChild(lexSrvcElement);
        	}
        }
        
        // EXTERNAL REFERENCES (arbitrary order)
        
        for (Entry<ExternalReference, String> entry : getExtRefIdParams.map.entrySet()) {
            ExternalReference er = entry.getKey();
            String id = entry.getValue();

            if (id != null && er != null) {
            	Element erElement = eafFactory.newExternalReference(id, er.getTypeString(), er.getValue());
        		annotDocument.appendChild(erElement);
        	}
        }
        
        return eafFactory.getDocumentElement();
    }

    /**
     * Common code for the DOM and the TemplateDOM to add the list of &lt;LANGUAGE>s
     * that are used in the collections of CVs and Tiers.
     * 
     * Since the tiers only store the short Ids, we need to find the corresponding
     * long Ids. They are supposed to be in the recent languages list though.
     * If that fails (it shouldn't) we fall back to the full official list.
     * 
     * @param storeOrder
     * @param eafFactory
     * @param annotDocument
     * @param conVocs
     */
	private static void addLanguages(List<Tier> storeOrder,
			EAF28 eafFactory, Element annotDocument,
			List<ControlledVocabulary> conVocs) {
		Map<String, LanguageRecord> defToLang = new HashMap<String, LanguageRecord>();
        Map<String, LanguageRecord> idToLang = new HashMap<String, LanguageRecord>();

        // Get languages from CVs
        for (ControlledVocabulary cv: conVocs) {
        	int nLangs = cv.getNumberOfLanguages();
        	for (int i = 0; i < nLangs; i++) {
        		String id = cv.getLanguageId(i);
        		String def = cv.getLongLanguageId(i);
        		String label = cv.getLanguageLabel(i);
        		final LanguageRecord lr = new LanguageRecord(id, def, label);
				defToLang.put(def, lr);
				idToLang.put(id, lr);
        	}
        }
        
        // Get languages from Tiers; stored only as (short) id.
        // We need to find the full record for the language.
        for (Tier tier : storeOrder) {
        	String id = tier.getLangRef(); // short Id
        	if (id != null && !id.isEmpty()) {
	        	// Try to look up in the languages that were collected already from the CVs
	        	LanguageRecord lr =  idToLang.get(id);
	        	if (lr == null) {
	        		// try to look up long Id just in case
	        		lr = defToLang.get(id);
	        	}
	        	// Otherwise, try the recent languages list
	        	if (lr == null) {
	        		LangInfo li = RecentLanguages.getInstance().getLanguageInfo(id);
	        		if (li != null) {
	        			lr = new LanguageRecord(li.getId(), li.getLongId(), li.getLabel());        			
	        		}
	        	}   
	        	// Otherwise, try the full official languages list
	        	if (lr == null) {
	        		LangInfo li = LanguageCollection.getLanguageInfo(id);
	        		if (li != null) {
	        			lr = new LanguageRecord(li.getId(), li.getLongId(), li.getLabel());        			
	        		}
	        	}   
	        	if (lr != null) {
	        		defToLang.put(lr.getDef(), lr);
					idToLang.put(lr.getId(), lr);
	        	}
        	}
        }

        // Now create <LANGUAGE> elements for all collected languages
        for (LanguageRecord lr : defToLang.values()) {
        	Element languageElement = eafFactory.newLanguage(lr.getId(), lr.getDef(), lr.getLabel());
        	annotDocument.appendChild(languageElement);
        }
	}
	
	/**
	 * A bit of state information for mapping ExternalReferences to ids for use in the
	 * saved file.
	 * <p>
	 * The Map could be a LinkedHashMap, if we care about printing the ExternalReferences
	 * in numerical order.
	 * 
	 * @author olasei
	 */
	private static class GetExtRefIdParams {
		private int extRefIndex;
		Map<ExternalReference, String> map;
		
		GetExtRefIdParams() {
			extRefIndex = 1;
			map = new HashMap<ExternalReference, String>();
		}
	}
	
	/**
	 * Gets an ID for a given ExternalReference. If identical ExternalReferences
	 * are passed more than once, it will produce the same ID every time.
	 * If passed null, it will return null.
	 * 
	 * @param params state object that keeps the context
	 * @param extRef
	 * @return
	 */
	private static String getExtRefId(GetExtRefIdParams params, ExternalReference extRef) {
		if (extRef == null || params == null) {
			return null;
		}
		
		String id = params.map.get(extRef);
		if (id == null) {
			id = "er" + params.extRefIndex++;
			params.map.put(extRef, id);
		}
		return id;
	}

    /**
     * Create the DOM tree containing only elements that need to be stored in
     * the  template; i.e. everything except timeorder, time slots,
     * annotations and media descriptors.
     *
     * @param theTranscription the Transcription to save (not null)
     * @param tierOrder the preferred ordering of the tiers
     *
     * @return the document element
     */
    public static Element createTemplateDOM(Transcription theTranscription,
        List<TierImpl> tierOrder, String path) {
        Map<String, Element> tierElements = new HashMap<String, Element>(); // for temporary storage of created tier Elements
                                                  //Hashtable annotationIds = new Hashtable(); // for temporary storage of generated annIds
        GetExtRefIdParams getExtRefIdParams = new GetExtRefIdParams();
        
        // For saving Lexicon Query Bundles:
        HashMap<String, LexiconQueryBundle2> lexRefs = new HashMap<String, LexiconQueryBundle2>();
        int lexRefIndex = 1;
        
        List<Locale> usedLocales = new ArrayList<Locale>(); // for storage of used locales

        TranscriptionImpl attisTr = (TranscriptionImpl) theTranscription;

        if (attisTr == null) {
            LOG.warning(
                "[[ASSERTION FAILED]] TranscriptionStore/storeTranscription: theTranscription is null");
        }

        EAF28 eafFactory = null;

        try {
            eafFactory = new EAF28();
        } catch (Exception ex) {
            ex.printStackTrace();
        }

        // ANNOTATION_DOCUMENT
        String dateString = getDate();

        String author = attisTr.getAuthor();

        //always set author to empty in template
        author = "";

        Element annotDocument = eafFactory.newAnnotationDocument(dateString,
                author, VERSION);
        eafFactory.appendChild(annotDocument);

        // LICENSEs - new in 2.8
        List<LicenseRecord> licenses = theTranscription.getLicenses();
        if (licenses != null) {
            for (LicenseRecord lr : licenses) {
                Element license = eafFactory.newLicense(lr);
                annotDocument.appendChild(license);
            }
        }

        // HEADER
        //always set header to empty in template
        
        Element header = eafFactory.newHeader("");
        annotDocument.appendChild(header);

        // TIME_ORDER element, empty. To satisfy the xsd. 
        Element timeOrderElement = eafFactory.newTimeOrder();
        annotDocument.appendChild(timeOrderElement);
        
        // TIERS
        //Vector tiers = attisTr.getTiers();
        // HS Nov 2009 always store in the order the tiers are in the transcription
        // for new files this means creation order
        List<Tier> storeOrder = new ArrayList<Tier>(attisTr.getTiers());
        /*
        Vector storeOrder = new Vector(tierOrder); // start with tiers in specified order
        Iterator tIter = tiers.iterator();

        while (tIter.hasNext()) { // add other tiers in document order

            Tier t = (Tier) tIter.next();

            if (!storeOrder.contains(t)) {
                storeOrder.add(t);
            }
        }
		*/
        Iterator<Tier> tierIter = storeOrder.iterator();

        while (tierIter.hasNext()) {
            TierImpl t = (TierImpl) tierIter.next();

            String id = t.getName();
            String participant = (String) t.getMetadataValue("PARTICIPANT");
            String annotator = (String) t.getMetadataValue("ANNOTATOR");
            String lingType = t.getLinguisticType().getLinguisticTypeName();

            if (lingType == null) {
                lingType = "not specified";
            }

            //Locale lang = (Locale) t.getMetadataValue("DEFAULT_LOCALE");
            Locale lang = t.getDefaultLocale();
            //  HS Feb 2010 allow null for Locale, use empty string internally
//            if (lang == null) {
//                lang = new Locale("not specified", "", "");
//            }

            // check is quick solution, TreeSet would do this but compareTo causes ClassCastException
            if (lang != null && !usedLocales.contains(lang)) {
                usedLocales.add(lang);
            }

            String parentName = null;

            if (t.getParentTier() != null) {
                parentName = t.getParentTier().getName();
            }

            String extRef = t.getExtRef();
            String langRef = t.getLangRef();

            Element tierElement = eafFactory.newTier(id, participant, annotator,
            		lingType, lang, parentName, extRef, langRef);
            annotDocument.appendChild(tierElement);

            tierElements.put(t.getName(), tierElement); // store for later use
        }

        // LINGUISTIC_TYPES
        List<LinguisticType> lTypes = attisTr.getLinguisticTypes();

        if (lTypes != null) {
            Iterator<LinguisticType> typeIter = lTypes.iterator();

            while (typeIter.hasNext()) {
                // HB, april 24, 2002: for the moment, just store lt name
                LinguisticType lt = typeIter.next();
                // may 2008 store and add external reference id ref
                String extRefId = null;
                if (lt.getDataCategory() != null && lt.getDataCategory().length() > 0) {
                	ExternalReferenceImpl eri = new ExternalReferenceImpl(lt.getDataCategory(), ExternalReference.ISO12620_DC_ID);
                	extRefId = getExtRefId(getExtRefIdParams, eri);
                }
                
                String stereotype = null;

                if (lt.hasConstraints()) {
                    stereotype = Constraint.stereoTypes[lt.getConstraints()
                                                          .getStereoType()];
                    stereotype = stereotype.replace(' ', '_');
                }

                // Create a LEXICON_REF if there is a Lexicon Query Bundle used.
                String lexRef = null;
                LexiconQueryBundle2 queryBundle = lt.getLexiconQueryBundle();
                if(queryBundle != null) {
                	lexRef = "lr" + lexRefIndex++;
                	lexRefs.put(lexRef, queryBundle);
                }
                
                Element typeElement = eafFactory.newLinguisticType(lt.getLinguisticTypeName(),
                        lt.isTimeAlignable(), lt.hasGraphicReferences(),
                        stereotype, lt.getControlledVocabularyName(), extRefId, 
                        lexRef);

                annotDocument.appendChild(typeElement);
            }
        }

        // LOCALES
        Iterator<Locale> locIter = usedLocales.iterator();

        while (locIter.hasNext()) {
            Locale l = locIter.next();
            Element locElement = eafFactory.newLocale(l);
            annotDocument.appendChild(locElement);
        }
        
        // <LANGUAGE>
        // Collect all used languages from the Controlled Vocabularies
        List<ControlledVocabulary> conVocs = attisTr.getControlledVocabularies();
        addLanguages(storeOrder, eafFactory, annotDocument, conVocs);
        
        // HB, 18 jul 02: for the moment manually add relevant Constraints
        // CONSTRAINTS
        Element timeSubdivision = eafFactory.newConstraint(Constraint.stereoTypes[Constraint.TIME_SUBDIVISION].replace(
                    ' ', '_'),
                "Time subdivision of parent annotation's time interval, no time gaps allowed within this interval");
        Element symbSubdivision = eafFactory.newConstraint(Constraint.stereoTypes[Constraint.SYMBOLIC_SUBDIVISION].replace(
                    ' ', '_'),
                "Symbolic subdivision of a parent annotation. Annotations refering to the same parent are ordered");
        Element symbAssociation = eafFactory.newConstraint(Constraint.stereoTypes[Constraint.SYMBOLIC_ASSOCIATION].replace(
                    ' ', '_'), "1-1 association with a parent annotation");
        Element includedIn = eafFactory.newConstraint(Constraint.stereoTypes[Constraint.INCLUDED_IN].replace(
                    ' ', '_'),
                "Time alignable annotations within the parent annotation's time interval, gaps are allowed");

        annotDocument.appendChild(timeSubdivision);
        annotDocument.appendChild(symbSubdivision);
        annotDocument.appendChild(symbAssociation);
        annotDocument.appendChild(includedIn);

      //CONTROLLED VOCABULARIES
        //Vector<ControlledVocabulary> conVocs = attisTr.getControlledVocabularies();

        if (conVocs.size() > 0) {
            Element cvElement;
            String extRefId = null;

            for (ControlledVocabulary cv : conVocs) {
                if (cv instanceof ExternalCV) {
        			// If cv is an external CV, save the cv with a EXT_REF and no entries in the EAF
        			// and save an ECV for caching
        			final ExternalReference externalRef = ((ExternalCV) cv).getExternalRef();
                	extRefId = getExtRefId(getExtRefIdParams, externalRef);
        			cvElement = eafFactory.newControlledVocabulary(cv.getName(),
        					null/*cv.getDescription()*/, extRefId);
        		} else {
        			cvElement = eafFactory.newControlledVocabulary(cv.getName(),
        					null/*cv.getDescription()*/);
        			
        			// <DESCRIPTION>s
        			int nLangs = cv.getNumberOfLanguages();
        			for (int i = 0; i < nLangs; i++) {
        				Element descElement = eafFactory.newDescription(
        						cv.getLanguageId(i),
        						cv.getDescription(i));
        				cvElement.appendChild(descElement);
        			}
        						
        			for (CVEntry entry : cv) {
        				final ExternalReference externalRef = entry.getExternalRef();
   	                	extRefId = getExtRefId(getExtRefIdParams, externalRef);

        				Element entryMLElement = eafFactory.newCVEntryML(
        						entry.getId(),
        						extRefId);
        				
        				// Add all non-empty values for the different languages.
        				for (int i = 0; i < nLangs; i++) {
        					String value = entry.getValue(i);
        					String descr = entry.getDescription(i);
        					if ((value != null && !value.isEmpty()) ||
        					    (descr != null && !descr.isEmpty())) {
	        					Element entryValueElement = eafFactory.newCVEntryValue(
	        							cv.getLanguageId(i),
	        							value,
	        							descr);
	        					entryMLElement.appendChild(entryValueElement);
        					}
        				}
        				
        				cvElement.appendChild(entryMLElement);
        			}
        		}


                annotDocument.appendChild(cvElement);
            }
        }

        // LEXICON SERVICES
        // First save the Lexicon Query Bundles
        Map<String, LexiconLink> savedLexiconLinks = new HashMap<String, LexiconLink>();
        for(String lexRef : lexRefs.keySet()) {
        	LexiconQueryBundle2 queryBundle = lexRefs.get(lexRef);
        	Element lexSrvcElement = eafFactory.newLexiconReference(lexRef, queryBundle);
        	savedLexiconLinks.put(queryBundle.getLinkName(), queryBundle.getLink());
        	annotDocument.appendChild(lexSrvcElement);
        }
        
        // Then save all Lexicon Links that were not in a Lexicon Query Bundle
        Map<String, LexiconLink> lexiconLinks = attisTr.getLexiconLinks();
        for(String linkName : lexiconLinks.keySet()) {
        	if(!savedLexiconLinks.containsKey(linkName)) {
        		Element lexSrvcElement = eafFactory.newLexiconLink("lr" + lexRefIndex++, lexiconLinks.get(linkName));
        		annotDocument.appendChild(lexSrvcElement);
        	}
        }
        
        // EXTERNAL REFERENCES

        for (Entry<ExternalReference, String> pair : getExtRefIdParams.map.entrySet()) {
            ExternalReference er = pair.getKey();
            String id = pair.getValue();

            if (id != null && er != null) {
            	Element erElement = eafFactory.newExternalReference(id, er.getTypeString(), er.getValue());
        		annotDocument.appendChild(erElement);
        	}
        }

        
        return eafFactory.getDocumentElement();
    }
   
    /**
     * Get a date of the form "2014-03-05T15:06:29+01:00".
     * @return
     */
    private static String getDate() {
        SimpleDateFormat dateFmt = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
        final Calendar calendar = Calendar.getInstance();
		String dateString = dateFmt.format(calendar.getTime());

		return addTimezone(calendar, dateString);
    }
    
    private static String addTimezone(Calendar calendar, String dateString) {
        int GMTOffsetInMinutes = calendar.getTimeZone().getRawOffset() / (60 * 1000);

        char sign = '+';

        if (GMTOffsetInMinutes < 0) {
            sign = '-';
            GMTOffsetInMinutes = -GMTOffsetInMinutes;
        }

        final int hours = GMTOffsetInMinutes / 60;
		final int minutes = GMTOffsetInMinutes % 60;

		String strOffset = String.format("%02d:%02d", hours, minutes);
        
        return dateString + sign + strOffset;
    }

    private static void save(Element documentElement, String path) throws IOException{
        LOG.info(path + " <----XML output\n");

        try {
            // test for errors
            if (("" + documentElement).length() == 0) {
                throw new IOException("Unable to save this file (zero length).");
                /*
                String txt = "Sorry: unable to save this file (zero length).";
                JOptionPane.showMessageDialog(null, txt, txt,
                    JOptionPane.ERROR_MESSAGE);

                //bla
                return;
                */
            }
            long beginTime = System.currentTimeMillis();
            //IoUtil.writeEncodedFile("UTF-8", path, documentElement);
            IoUtil.writeEncodedEAFFile("UTF-8", path, documentElement);
            
            if (debug) {
            	System.out.println("Saving file took: " + 
            			(System.currentTimeMillis() - beginTime) + " ms");	
            }
        } catch (Exception eee) {
            throw new IOException("Unable to save this file: " + eee.getMessage());
            /*
            String txt = "Sorry: unable to save this file. (" +
                eee.getMessage() + ")";
            JOptionPane.showMessageDialog(null, txt, txt,
                JOptionPane.ERROR_MESSAGE);
                */
        }
    }
}
